//
//  Copyright (c) 2021 Open Whisper Systems. All rights reserved.
//

import Foundation
import PromiseKit
import UIKit

protocol GroupAttributesViewControllerDelegate: AnyObject {
    func groupAttributesDidUpdate()
}

// MARK: -

class GroupAttributesViewController: OWSTableViewController2 {

    public enum EditAction {
        case none
        case avatar
    }

    private weak var attributesDelegate: GroupAttributesViewControllerDelegate?

    private let groupThread: TSGroupThread

    private let helper: GroupAttributesEditorHelper

    private var editAction: EditAction?

    private var hasUnsavedChanges: Bool {
        return helper.hasUnsavedChanges
    }

    public required init(groupThread: TSGroupThread,
                         editAction: EditAction,
                         delegate: GroupAttributesViewControllerDelegate) {
        self.groupThread = groupThread
        self.editAction = editAction
        self.attributesDelegate = delegate

        self.helper = GroupAttributesEditorHelper(groupModel: groupThread.groupModel)

        super.init()
    }

    // MARK: - View Lifecycle

    @objc
    public override func viewDidLoad() {
        super.viewDidLoad()

        view.backgroundColor = Theme.backgroundColor

        title = NSLocalizedString("EDIT_GROUP_DEFAULT_TITLE", comment: "The navbar title for the 'update group' view.")

        defaultSeparatorInsetLeading = Self.cellHInnerMargin + 24 + OWSTableItem.iconSpacing

        helper.delegate = self
        helper.buildContents(avatarViewHelperDelegate: self)

        updateTableContents()
    }

    func updateTableContents() {
        let contents = OWSTableContents()

        let avatarSection = OWSTableSection()
        avatarSection.hasBackground = false
        avatarSection.add(.init(
            customCellBlock: { [weak self] in
                let cell = OWSTableItem.newCell()
                cell.selectionStyle = .none
                guard let self = self else { return cell }
                cell.contentView.addSubview(self.helper.avatarWrapper)
                self.helper.avatarWrapper.autoPinHeightToSuperviewMargins()
                self.helper.avatarWrapper.autoHCenterInSuperview()
                return cell
            },
            actionBlock: { [weak self] in
                self?.didTapAvatarView()
            }
        ))
        contents.addSection(avatarSection)

        if FeatureFlags.groupDescriptionEditing {
            let nameAndDescriptionSection = OWSTableSection()
            nameAndDescriptionSection.add(.disclosureItem(
                icon: .settingsAddToGroup,
                name: helper.groupNameCurrent ?? NSLocalizedString(
                    "GROUP_NAME_VIEW_TITLE",
                    comment: "Title for the group name view."
                ),
                accessibilityIdentifier: UIView.accessibilityIdentifier(in: self, name: "group_name"),
                actionBlock: { [weak self] in
                    guard let self = self else { return }
                    let vc = GroupNameViewController(
                        groupModel: self.groupThread.groupModel,
                        groupNameCurrent: self.helper.groupNameCurrent
                    )
                    vc.nameDelegate = self
                    self.presentFormSheet(OWSNavigationController(rootViewController: vc), animated: true)
                }
            ))
            nameAndDescriptionSection.add(.disclosureItem(
                icon: .compose24,
                name: helper.groupDescriptionCurrent ?? NSLocalizedString(
                    "GROUP_DESCRIPTION_VIEW_TITLE",
                    comment: "Title for the group description view."
                ),
                maxNameLines: 2,
                accessibilityIdentifier: UIView.accessibilityIdentifier(in: self, name: "group_description"),
                actionBlock: { [weak self] in
                    guard let self = self else { return }
                    let vc = GroupDescriptionViewController(
                        groupModel: self.groupThread.groupModel,
                        groupDescriptionCurrent: self.helper.groupDescriptionCurrent,
                        options: .editable
                    )
                    vc.descriptionDelegate = self
                    self.presentFormSheet(OWSNavigationController(rootViewController: vc), animated: true)
                }
            ))
            contents.addSection(nameAndDescriptionSection)
        } else {
            let nameTextField = helper.nameTextField
            let nameSection = OWSTableSection()
            nameSection.add(.init(
                customCellBlock: {
                    let cell = OWSTableItem.newCell()
                    cell.selectionStyle = .none

                    nameTextField.font = .ows_dynamicTypeBodyClamped
                    nameTextField.textColor = Theme.primaryTextColor

                    cell.contentView.addSubview(nameTextField)
                    nameTextField.autoPinEdgesToSuperviewMargins()

                    return cell
                },
                actionBlock: {
                    nameTextField.becomeFirstResponder()
                }
            ))
            contents.addSection(nameSection)
        }

        self.contents = contents
    }

    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)

        if let editAction = self.editAction {
            switch editAction {
            case .none:
                break
            case .avatar:
                helper.showAvatarUI()
            }
            self.editAction = nil
        }
    }

    // MARK: -

    fileprivate func updateNavbar() {
        if helper.hasUnsavedChanges {
            navigationItem.rightBarButtonItem = UIBarButtonItem(
                title: CommonStrings.setButton,
                style: .done,
                target: self,
                action: #selector(setButtonPressed),
                accessibilityIdentifier: "set_button"
            )
        } else {
            navigationItem.rightBarButtonItem = nil
        }
    }

    @objc
    func setButtonPressed() {
        updateGroupThreadAndDismiss()
    }

    // MARK: - Events

    @objc
    func didTapAvatarView() {
        helper.didTapAvatarView()
    }
}

// MARK: -

extension GroupAttributesViewController: AvatarViewHelperDelegate {
    func avatarActionSheetTitle() -> String? {
        return NSLocalizedString("NEW_GROUP_ADD_PHOTO_ACTION", comment: "Action Sheet title prompting the user for a group avatar")
    }

    func avatarDidChange(_ image: UIImage) {
        helper.setAvatarImage(image)
    }

    func fromViewController() -> UIViewController {
        return self
    }

    func hasClearAvatarAction() -> Bool {
        return helper.avatarCurrent != nil
    }

    func clearAvatar() {
        helper.setAvatarImage(nil)
    }

    func clearAvatarActionLabel() -> String {
        return NSLocalizedString("EDIT_GROUP_CLEAR_AVATAR", comment: "The 'clear avatar' button in the 'edit group' view.")
    }
}

// MARK: -

extension GroupAttributesViewController: OWSNavigationView {

    public func shouldCancelNavigationBack() -> Bool {
        let result = hasUnsavedChanges
        if result {
            GroupAttributesViewController.showUnsavedGroupChangesActionSheet(from: self,
                                                                             saveBlock: {
                                                                                self.updateGroupThreadAndDismiss()
            }, discardBlock: {
                self.navigationController?.popViewController(animated: true)
            })
        }
        return result
    }

    @objc
    public static func showUnsavedGroupChangesActionSheet(from fromViewController: UIViewController,
                                                     saveBlock: @escaping () -> Void,
                                                     discardBlock: @escaping () -> Void) {
        let actionSheet = ActionSheetController(title: NSLocalizedString("EDIT_GROUP_VIEW_UNSAVED_CHANGES_TITLE",
                                                                         comment: "The alert title if user tries to exit update group view without saving changes."),
                                                message: NSLocalizedString("EDIT_GROUP_VIEW_UNSAVED_CHANGES_MESSAGE",
                                                                          comment: "The alert message if user tries to exit update group view without saving changes."))
        actionSheet.addAction(ActionSheetAction(title: CommonStrings.saveButton,
                                                accessibilityIdentifier: UIView.accessibilityIdentifier(in: fromViewController, name: "save"),
                                                style: .default) { _ in
                                                    saveBlock()
        })
        actionSheet.addAction(ActionSheetAction(title: NSLocalizedString("ALERT_DONT_SAVE",
                                                                         comment: "The label for the 'don't save' button in action sheets."),
                                                accessibilityIdentifier: UIView.accessibilityIdentifier(in: fromViewController, name: "dont_save"),
                                                style: .destructive) { _ in
                                                    discardBlock()
        })
        fromViewController.presentActionSheet(actionSheet)
    }
}

// MARK: -

private extension GroupAttributesViewController {
    func updateGroupThreadAndDismiss() {
        helper.updateGroupIfNecessary(fromViewController: self) { [weak self] in
            self?.attributesDelegate?.groupAttributesDidUpdate()
            self?.navigationController?.popViewController(animated: true)
        }
    }
}

// MARK: -

extension GroupAttributesViewController: GroupAttributesEditorHelperDelegate {
    func groupAttributesEditorContentsDidChange() {
        updateNavbar()
    }

    func groupAttributesEditorSelectionDidChange() {}
}

extension GroupAttributesViewController: GroupDescriptionViewControllerDelegate {
    func groupDescriptionViewControllerDidComplete(groupDescription: String?) {
        helper.groupDescriptionCurrent = groupDescription
        updateNavbar()
        updateTableContents()
    }
}

extension GroupAttributesViewController: GroupNameViewControllerDelegate {
    func groupNameViewControllerDidComplete(groupName: String?) {
        helper.groupNameCurrent = groupName
        updateNavbar()
        updateTableContents()
    }
}
